/* $Id: cmdline.c 1.5 2000/12/21 11:56:11 satomii Exp $ */
/*----------------------------------------------------------------------*/
/*		cmdline.c														*/
/*			designed		by satomii		00/12/15					*/
/*			copyright (c) 1999-2000, satomii. all rights reserved.		*/
/*----------------------------------------------------------------------*/
/*
 *	parses command-line options.
 *
 *	this program is free software; you can redistribute and/or modify it
 *	under the terms of the GNU General Public License either version 2 or
 *	later, as published by the Free Software Foundation.
 *
 *	this program is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *	General Public License for more details.
 */
#include <windows.h>
#include <tchar.h>
#include "cmdline.h"

#define _TCHAR(c)						((TCHAR)(c))
#define OptionID(optfmt)				((optfmt)->id)

#define IsShortOption(optfmt, optch)	\
		_tcsrchr((optfmt)->short_option, optch)

#define IsParamRequired(optfmt)			(_TCHAR('=') == (optfmt)->type)
#define IsParamOptional(optfmt)			(_TCHAR(':') == (optfmt)->type)
#define CanTakeParam(optfmt)			\
		(IsParamRequired(optfmt) || IsParamOptional(optfmt))


static int ShiftArgv(LPCTSTR *argv, int nFrom, int nTo)
{
	LPCTSTR lpstr;
	int i;

	lpstr = *(argv + nTo);
	for (i = nTo; nFrom < i; i--) *(argv + i) = *(argv + i - 1);
	*(argv + nFrom) = lpstr;
	return(nFrom + 1);
}

static int ParseShortOption(int index, int argc, LPCTSTR *argv,
		int nFormat, const OPTIONFORMAT FAR *lpFormat,
		PARSECMDCALLBACK pfnCallback, LPARAM lParam)
{
	LPCTSTR lpstr;
	int optid, res, i;

	/*
	 *	find the first option from the caller-supplied list.
	 */
	lpstr = *(argv + index) + 1;
	if (!*lpstr)
		return((pfnCallback)(PARSECMD_UNKNOWN_OPTION,
				0, NULL, lParam, *(argv + index), NULL));

	for (optid = 0; optid < nFormat; optid++) {
		if (!IsShortOption(lpFormat + optid, *lpstr)) continue;
		if (CanTakeParam(lpFormat + optid)) break;

		/*
		 *	if the option does not take an argument, the remaining part of
		 *	of this parameter cannot contain any option with an argument,
		 */
		while (*(++lpstr)) {
			for (i = 0; i < nFormat; i++) {
				if (IsShortOption(lpFormat + i, *lpstr)
						&& CanTakeParam(lpFormat + i))
					return((pfnCallback)(PARSECMD_INVALID_FORMAT,
							0, NULL, lParam, *(argv + index), NULL));
			}
		}
		break;
	}

	/*
	 *	if the first character indicates a valid or invalid option which
	 *	takes no parameter, all subsequent characters are also assumed as
	 *	option switches without parameters.
	 */
	if (nFormat <= optid || !CanTakeParam(lpFormat + optid)) {
		lpstr = *(argv + index) + 1;

		do {
#if !defined UNICODE && !defined _UNICODE
			if (IsDBCSLeadByte(*lpstr)) {
				res = (pfnCallback)(PARSECMD_UNKNOWN_OPTION,
						MAKEWORD(*lpstr, *(lpstr + 1)),
						NULL, lParam, *(argv + index), NULL);
				lpstr++;
			} else {
#endif	/* !UNICODE && !_UNICODE */
			for (optid = 0; optid < nFormat; optid++) {
				if (IsShortOption(lpFormat + optid, *lpstr)) break;
			}
			if (nFormat <= optid)
				res = (pfnCallback)(PARSECMD_UNKNOWN_OPTION,
						*lpstr, NULL, lParam, *(argv + index), NULL);
			else res = (pfnCallback)(PARSECMD_NOERROR,
					OptionID(lpFormat + optid), NULL, lParam,
					*(argv + index), NULL);
#if !defined UNICODE && !defined _UNICODE
			}
#endif	/* !UNICODE && !_UNICODE */
			if (PARSECMD_ABORT == res
					|| (lpstr == *(argv + index) + 1
					&& PARSECMD_ACCEPT != res))
				return(res);
		} while (*(++lpstr));

		return(PARSECMD_ACCEPT);
	}

	/*
	 *	otherwise notify the caller of this option and its parameter.
	 */
	lpFormat += optid;

	if (*(++lpstr))
		return((pfnCallback)(PARSECMD_NOERROR,
				OptionID(lpFormat), lpstr, lParam, *(argv + index), NULL));

	if (IsParamOptional(lpFormat))
		return((pfnCallback)(PARSECMD_NOERROR,
				OptionID(lpFormat), NULL, lParam, *(argv + index), NULL));

	if (index + 1 < argc) {
		res = (pfnCallback)(PARSECMD_NOERROR,
				OptionID(lpFormat), *(argv + index + 1), lParam,
				*(argv + index), *(argv + index + 1));
		if (PARSECMD_ACCEPT == res) return(MAKELONG(res, 1));
		return(res);
	}

	return((pfnCallback)(PARSECMD_PARAM_REQUIRED,
			OptionID(lpFormat), NULL, lParam, *(argv + index), NULL));
}

static int ParseLongOption(int index, int argc, LPCTSTR *argv,
		int nFormat, const OPTIONFORMAT FAR *lpFormat,
		PARSECMDCALLBACK pfnCallback, LPARAM lParam)
{
	LPCTSTR lpstr, lpszParam = NULL;
	int optid, len;

	/*
	 *	find the long option from the caller-supplied list.
	 */
	lpstr = *(argv + index) + 2;
	for (optid = 0; optid < nFormat; optid++) {
		len = lstrlen((lpFormat + optid)->long_option);
		if (_tcsncmp((lpFormat + optid)->long_option, lpstr, len)) continue;
		if (!*(lpstr + len)) break;
		if (_TCHAR('=') == *(lpstr + len)) {
			lpszParam = lpstr + len + 1;
			break;
		}
	}
	if (nFormat <= optid)
		return((pfnCallback)(PARSECMD_UNKNOWN_OPTION,
				0, NULL, lParam, *(argv + index), NULL));
	lpFormat += optid;

	/*
	 *	check whether this option takes a parameter.
	 */
	if (IsParamOptional(lpFormat))
		return((pfnCallback)(PARSECMD_NOERROR,
				OptionID(lpFormat), lpszParam, lParam,
				*(argv + index), NULL));

	if (!IsParamRequired(lpFormat))
		return((pfnCallback)(lpszParam ?
				PARSECMD_PARAM_NOT_ALLOWED : PARSECMD_NOERROR,
				OptionID(lpFormat), lpszParam, lParam,
				*(argv + index), NULL));

	if (lpszParam)
		return((pfnCallback)(PARSECMD_NOERROR,
				OptionID(lpFormat), lpszParam, lParam,
				*(argv + index), NULL));

	if (index + 1 < argc) {
		int res = (pfnCallback)(PARSECMD_NOERROR,
				OptionID(lpFormat), *(argv + index + 1), lParam,
				*(argv + index), *(argv + index + 1));
		if (PARSECMD_ACCEPT == res) return(MAKELONG(res, 1));
		return(res);
	}

	return((pfnCallback)(PARSECMD_PARAM_REQUIRED,
			OptionID(lpFormat), NULL, lParam, *(argv + index), NULL));
}

int ParseCommandLine(int argc, LPCTSTR *argv,
		int nFormat, const OPTIONFORMAT FAR *lpFormat,
		PARSECMDCALLBACK pfnCallback, LPARAM lParam)
{
	LPCTSTR lpstr;
	int nParam = argc;
	int res, i;

#ifdef _CONSOLE
	/* disable control events during this process. */
	HANDLE hStdIn;
	DWORD dwMode = 0;

	hStdIn = GetStdHandle(STD_INPUT_HANDLE);
	if (INVALID_HANDLE_VALUE != hStdIn && GetConsoleMode(hStdIn, &dwMode)) {
		if (!(dwMode & ENABLE_LINE_INPUT)
				|| !SetConsoleMode(hStdIn, dwMode | ENABLE_LINE_INPUT))
			hStdIn = NULL;
	} else {
		hStdIn = NULL;
	}
#endif	/* _CONSOLE */

	for (i = 0; i < argc; i++) {
		lpstr = *(argv + i);

		if (_TCHAR('-') == *lpstr && _TCHAR('-') == *(lpstr + 1)) {
			/* may be a long option. */
			lpstr += 2;
			if (!*lpstr) {
				/* "--" indicates the end of option switches. */
				if (nParam < i) nParam = ShiftArgv(argv, nParam, i);
				else nParam = i + 1;
				break;
			}
			res = ParseLongOption(i, argc, argv, nFormat, lpFormat,
					pfnCallback, lParam);
		} else if (_TCHAR('-') == *lpstr || _TCHAR('/') == *lpstr) {
			/* may be a short option. */
			res = ParseShortOption(i, argc, argv, nFormat, lpFormat,
					pfnCallback, lParam);
		} else {
			/* not an option switch. */
			res = PARSECMD_IGNORE;
		}

		if (PARSECMD_ABORT == res) {
			nParam = -1;
			break;
		}
		if (PARSECMD_ACCEPT == LOWORD(res)) {
			i += HIWORD(res);
			if (nParam < i) nParam = ShiftArgv(argv, nParam, i);
		} else {
			if (i < nParam) nParam = i;
		}
	}

#ifdef _CONSOLE
	/* restore previous control mode. */
	if (hStdIn) SetConsoleMode(hStdIn, dwMode);
#endif	/* _CONSOLE */
	return(nParam);
}
