/* $Id: console.c 1.5 2000/12/21 11:56:11 satomii Exp $ */
/*----------------------------------------------------------------------*/
/*		console.c														*/
/*			designed		by satomii		00/12/08					*/
/*			copyright (c) 1999-2000, satomii. all rights reserved.		*/
/*----------------------------------------------------------------------*/
/*
 *	this program is free software; you can redistribute and/or modify it
 *	under the terms of the GNU General Public License either version 2 or
 *	later, as published by the Free Software Foundation.
 *
 *	this program is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *	General Public License for more details.
 */
#include <windows.h>
#include "doc2txt.h"


/*------------------------------------------------------------------------
	console output
------------------------------------------------------------------------*/

DWORD ConPrint(DWORD dwHandle, LPCTSTR lpszText)
{
	HANDLE hStd;
	DWORD dw = 0;

	hStd = GetStdHandle(dwHandle);
	if (INVALID_HANDLE_VALUE == hStd) return(0);
	WriteFile(hStd, lpszText, lstrlen(lpszText), &dw, NULL);
#ifdef _DEBUG
	OutputDebugString(lpszText);
#endif	/* _DEBUG */
	return(dw);
}

DWORD ConPrintFormat(DWORD dwHandle, LPCTSTR lpszFmt, ...)
{
	HANDLE hStd;
	LPTSTR lpstr;
	DWORD dw = 0;
	va_list va;

	hStd = GetStdHandle(dwHandle);
	if (INVALID_HANDLE_VALUE == hStd) return(0);

	va_start(va, lpszFmt);
	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
			FORMAT_MESSAGE_FROM_STRING,
			lpszFmt, 0, 0, (LPSTR)&lpstr, 0, &va);
	va_end(va);

	if (lpstr) {
		WriteFile(hStd, lpstr, lstrlen(lpstr), &dw, NULL);
#ifdef _DEBUG
		OutputDebugString(lpstr);
#endif	/* _DEBUG */
		LocalFree(lpstr);
	} else {
		WriteFile(hStd, lpszFmt, lstrlen(lpszFmt), &dw, NULL);
#ifdef _DEBUG
		OutputDebugString(lpszFmt);
#endif	/* _DEBUG */
	}

	/* do not close the standard handle, it is not necessary. */
	return(dw);
}

DWORD ConPrintFile(DWORD dwHandle, LPCTSTR lpszPath)
{
	HANDLE hStd, hFile;
	DWORD dwLen = 0, dw;
	BYTE b;

	hStd = GetStdHandle(dwHandle);
	if (INVALID_HANDLE_VALUE == hStd) return(0);

	hFile = CreateFile(lpszPath, GENERIC_READ, FILE_SHARE_READ,
			NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (INVALID_HANDLE_VALUE == hFile) return(0);

	while (!bAbortOperation) {
		if (!ReadFile(hFile, &b, sizeof(b), &dw, NULL)
				|| !dw
				|| !WriteFile(hStd, &b, dw, &dw, NULL))
			break;
		dwLen += dw;
	}

	dw = GetLastError();
	CloseHandle(hFile);
	SetLastError(dw);
	return(dwLen);
}


#if 0

/*------------------------------------------------------------------------
	console input
------------------------------------------------------------------------*/

int ConWaitCharInput(void)
{
	HANDLE hStdIn;
	INPUT_RECORD ir;
	DWORD dwRead;
	int ch = 0;

	hStdIn = GetStdHandle(STDIN);
	if (INVALID_HANDLE_VALUE == hStdIn) return(0);

	while (!bAbortOperation) {
		if (!ReadConsoleInputA(hStdIn, &ir, 1, &dwRead)) {
			ch = 0;
			break;
		}
		if (KEY_EVENT != ir.EventType
				|| !(ir.Event.KeyEvent.bKeyDown)
				|| !(ir.Event.KeyEvent.uChar.AsciiChar)
				|| (ir.Event.KeyEvent.dwControlKeyState
					& (LEFT_CTRL_PRESSED | RIGHT_CTRL_PRESSED
					| LEFT_ALT_PRESSED | RIGHT_ALT_PRESSED)))
			continue;
		if (ch) {
			ch = (int)MAKEWORD(ch, ir.Event.KeyEvent.uChar.AsciiChar);
			break;
		} else if (IsDBCSLeadByte(ir.Event.KeyEvent.uChar.AsciiChar)) {
			ch = ir.Event.KeyEvent.uChar.AsciiChar;
		} else {
			ch = ir.Event.KeyEvent.wVirtualKeyCode;
			break;
		}
	}

	return(ch);
}

DWORD ConGetLine(LPTSTR lpszText, DWORD cchText)
{
	HANDLE hStdIn;
	DWORD dwRead;

	hStdIn = GetStdHandle(STDIN);
	if (INVALID_HANDLE_VALUE == hStdIn) return(0);
	if (!ReadFile(hStdIn, lpszText, cchText, &dwRead, NULL)) return(0);
	return(dwRead);
}

#endif	/* 0 */
