/* $Id: conv.c 1.5.0.1 2000/12/22 00:26:50 satomii Exp $ */
/*----------------------------------------------------------------------*/
/*		conv.c															*/
/*			designed		by satomii		00/12/17					*/
/*			copyright (c) 1999-2000, satomii. all rights reserved.		*/
/*----------------------------------------------------------------------*/
/*
 *	this program is free software; you can redistribute and/or modify it
 *	under the terms of the GNU General Public License either version 2 or
 *	later, as published by the Free Software Foundation.
 *
 *	this program is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *	General Public License for more details.
 */
#include <windows.h>
#include "doc2txt.h"
#include "msgtbl.h"

#define QUIET					(lpParam->dwFlags & FLAG_QUIET)
#define DEBUG					(lpParam->dwFlags & FLAG_DEBUG)


/*------------------------------------------------------------------------
	output filename creation
------------------------------------------------------------------------*/

BOOL MakeOutputPath(LPCTSTR lpszFrom,
		LPTSTR lpszTo, DWORD cchTo, LPCMDPARAM lpParam)
{
	LPTSTR lpstr;
	HANDLE hFind;
	WIN32_FIND_DATA wfd;
	TCHAR szExt[_MAX_EXT];
	DWORD dwLen;
	int i = 0;

	/*
	 *	copy the output directory and original filename.
	 */
	if (lpParam->dwFlags & FLAG_ORIGPATH) {
		dwLen = lstrlen(lpszFrom);
		if (dwLen < cchTo) lstrcpy(lpszTo, lpszFrom);
	} else {
		lpstr = _tcsrchr(lpszFrom, _TCHAR('\\'));
		if (lpstr) lpstr++;
		else lpstr = (LPTSTR)lpszFrom;
		dwLen = lpParam->dwOutDirLen + lstrlen(lpstr);
		if (dwLen < cchTo) {
			lstrcpy(lpszTo, lpParam->szOutDir);
			lstrcat(lpszTo, lpstr);
		}
	}
	if (cchTo <= dwLen) {
		Gripe(MSG_W_TOO_LONG_FILENAME, 0);
		return(FALSE);
	}

	/*
	 *	change or append the extension.
	 */
	if (lpParam->dwFlags & FLAG_APPEND_EXTENSION) {
		lpstr = lpszTo + dwLen;
	} else {
		lpstr = _tcsrchr(lpszTo, _TCHAR('.'));
		if (lpstr) {
			*lpstr = 0;
			dwLen = (DWORD)(lpstr - lpszTo);
		} else {
			lpstr = lpszTo + dwLen;
		}
	}
	dwLen += lpParam->dwOutExtLen;
	if (cchTo <= dwLen) {
		Gripe(MSG_W_TOO_LONG_FILENAME, 0);
		return(FALSE);
	}
	lstrcpy(lpstr, lpParam->szOutExt);

	/*
	 *	check whether the output file already exists.
	 */
	hFind = FindFirstFile(lpszTo, &wfd);
	if (INVALID_HANDLE_VALUE == hFind) return(TRUE);
	FindClose(hFind);

	if (lpParam->dwFlags & FLAG_OVERWRITE) return(TRUE);
	if (lpParam->dwFlags & FLAG_SKIP_EXISTING) {
		Gripe(MSG_W_OUTFILE_EXISTS, 0, lpszFrom, lpszTo);
		return(FALSE);
	}
	if (cchTo <= dwLen + 5) {
		Gripe(MSG_W_TOO_LONG_FILENAME, 0, lpszFrom);
		return(FALSE);
	}

	/*
	 *	try to make unique filename. this is the default behavior.
	 */
	TRACE1(_T("Output file \"%1\" exists, creating a unique name.\n"),
			lpszTo);

	lpstr = _tcschr(lpszTo, _TCHAR('.'));
	if (lpstr) {
		if (TCCH(szExt) <= lstrlen(lpstr))
			Gripe(MSG_W_TOO_LONG_FILENAME, 0);
		lstrcpy(szExt, lpstr);
	} else {
		lpstr = lpszTo + dwLen;
		szExt[0] = 0;
	}

	while (!bAbortOperation) {
		if (100 <= i) {
			Gripe(MSG_W_OUTFILE_FULL, 0);
			break;
		}
		wsprintf(lpstr, _T(" (%d)%s"), ++i, szExt);
		hFind = FindFirstFile(lpszTo, &wfd);
		if (INVALID_HANDLE_VALUE == hFind) return(TRUE);
		FindClose(hFind);
	}

	return(FALSE);
}


/*------------------------------------------------------------------------
	directory deletion
------------------------------------------------------------------------*/

BOOL DeleteDirectory(LPCSTR lpszPath)
{
	SHFILEOPSTRUCT fos;

	ZeroMemory(&fos, sizeof(fos));
	fos.wFunc = FO_DELETE;
	fos.pFrom = lpszPath;
	fos.fFlags = FOF_NOCONFIRMATION | FOF_NOERRORUI | FOF_SILENT;
	if (!SHFileOperation(&fos)) return(TRUE);
	return(FALSE);
}


/*------------------------------------------------------------------------
	Word initialization
------------------------------------------------------------------------*/

BOOL WordInitialize(LPCMDPARAM lpParam)
{
	HRESULT hr;
	LPUNKNOWN lpUnk;
	LPWSTR lpwszProgID = NULL;
	CLSID clsid;
	VARIANT varRet;

	TRACE0(_T("Initializing Word Application object.\n"));

	hr = OleInitialize(NULL);
	if (!SUCCEEDED(hr)) {
		Gripe(MSG_E_OLE_INITIALIZE, hr);
		return(FALSE);
	}

	if (lpParam->lpszProgID) {
#if defined UNICODE || defined _UNICODE
		lpwszProgID = lpParam->lpszProgID;
#else	/* !UNICODE && !_UNICODE */
		lpwszProgID = ToWideChar(lpParam->lpszProgID);
#endif	/* UNICODE || _UNICODE */
		if (!lpwszProgID) {
			Gripe(MSG_E_OUTOFMEMORY, 0);
			return(FALSE);
		}
	}

	hr = CLSIDFromProgID(lpwszProgID ?
			lpwszProgID : OLESTR("Word.Application"), &clsid);
#if !defined UNICODE && !defined _UNICODE
	if (lpwszProgID) LocalFree(lpwszProgID);
#endif	/* !UNICODE && !_UNICODE */

	if (SUCCEEDED(hr)) {
		hr = CoCreateInstance(&clsid,
				NULL, CLSCTX_SERVER, &IID_IUnknown, (LPVOID *)&lpUnk);
		if (SUCCEEDED(hr)) {
			hr = lpUnk->lpVtbl->QueryInterface(lpUnk,
					&IID_IDispatch, (LPVOID *)&(lpParam->lpApp));
			lpUnk->lpVtbl->Release(lpUnk);
		}
	}
	if (!SUCCEEDED(hr)) {
		if (!(lpParam->lpApp)) OleUninitialize();
		if (!(lpParam->dwFlags & FLAG_QUIET)) PRINT(STDOUT, _T("\n"));
		Gripe(MSG_E_WORD_INITIALIZE, hr);
		return(FALSE);
	}

	/* documents = Application.Documents */
	hr = DispatchInvoke(lpParam->lpApp, DISPATCH_PROPERTYGET,
			OLESTR("Documents"), &varRet, 0, NULL);
	if (!SUCCEEDED(hr)) {
		if (!(lpParam->dwFlags & FLAG_QUIET)) PRINT(STDOUT, _T("\n"));
		Gripe(MSG_E_WORD_GETDOCUMENT, hr);
		return(FALSE);
	}
	lpParam->lpDocs = V_DISPATCH(&varRet);
	lpParam->lpDocs->lpVtbl->AddRef(lpParam->lpDocs);
	VariantClear(&varRet);

	return(TRUE);
}


/*------------------------------------------------------------------------
	comment deletion
------------------------------------------------------------------------*/

HRESULT WordRemoveComment(LPDISPATCH lpDoc, LPCMDPARAM lpParam)
{
	HRESULT hr;
	VARIANT varRet;
	LPDISPATCH lpCollection, lpItem;
	LONG lCount, l;

	TRACE0(_T("Trying to remove comments.\n"));
	VariantInit(&varRet);

	/*
	 *	obtain the Comments collection object interface.
	 */
	hr = DispatchInvoke(lpDoc, DISPATCH_PROPERTYGET,
			OLESTR("Comments"), &varRet, 0, NULL);
	if (!SUCCEEDED(hr)) {
		TRACE1(_T("Failed to obtain Comments object (0x%1!08x!).\n"), hr);
		return(hr);
	}
	lpCollection = V_DISPATCH(&varRet);
	lpCollection->lpVtbl->AddRef(lpCollection);
	VariantClear(&varRet);

	/*
	 *	get the number of items in the collection.
	 */
	hr = DispatchInvoke(lpCollection, DISPATCH_PROPERTYGET,
			OLESTR("Count"), &varRet, 0, NULL);
	if (!SUCCEEDED(hr)) {
		TRACE1(_T("Comments.Count failed with 0x%1!08x!.\n"), hr);
		lpCollection->lpVtbl->Release(lpCollection);
		return(hr);
	}
	lCount = V_I4(&varRet);
	VariantClear(&varRet);
	TRACE1(_T("Comments.Count = %1!d!\n"), lCount);

	/*
	 *	try to remove all collection items.
	 */
	for (l = lCount; 0 < l; l--) {
		if (bAbortOperation) break;

		/* Comment = Comments->Item(index) */
		hr = DispatchInvoke(lpCollection, DISPATCH_METHOD,
				OLESTR("Item"), &varRet, 1, _T("i"), l);
		if (!SUCCEEDED(hr)) {
			TRACE2(_T("Comments->Item(%1!d!) failed with 0x%2!08x!.\n"),
					l, hr);
			continue;
		}

		lpItem = V_DISPATCH(&varRet);
		lpItem->lpVtbl->AddRef(lpItem);
		VariantClear(&varRet);

		/* Comment->Delete() */
		hr = DispatchInvoke(lpItem, DISPATCH_METHOD,
				OLESTR("Delete"), &varRet, 0, NULL);
		if (SUCCEEDED(hr)) {
			VariantClear(&varRet);
		} else {
			TRACE2(_T("Item(%1!d!)->Delete() failed with 0x%2!08x!.\n"),
					l, hr);
		}

		lpItem->lpVtbl->Release(lpItem);
	}

	/*
	 *	clean up.
	 */
#ifdef _DEBUG
	{
	HRESULT hrRel = lpCollection->lpVtbl->Release(lpCollection);
	PRINTF(STDERR, _T("debug: Comments->Release() returned %1!d!.\n"), hrRel);
	}
#else	/* !_DEBUG */
	lpCollection->lpVtbl->Release(lpCollection);
#endif	/* _DEBUG */

	if (bAbortOperation) return(E_ABORT);
	return(S_OK);
}


/*------------------------------------------------------------------------
	style deletion
------------------------------------------------------------------------*/

HRESULT WordRemoveStyle(LPDISPATCH lpDoc, LPCMDPARAM lpParam)
{
	HRESULT hr;
	VARIANT varRet;
	LPDISPATCH lpCollection, lpItem;
	LONG lCount, l;
	BOOL bBuiltIn;

	TRACE0(_T("Trying to remove styles.\n"));
	VariantInit(&varRet);

	/*
	 *	obtain the Styles collectio object interface.
	 */
	hr = DispatchInvoke(lpDoc, DISPATCH_PROPERTYGET,
			OLESTR("Styles"), &varRet, 0, NULL);
	if (!SUCCEEDED(hr)) {
		TRACE1(_T("Document.Styles failed with 0x%1!08x!.\n"), hr);
		return(hr);
	}
	lpCollection = V_DISPATCH(&varRet);
	lpCollection->lpVtbl->AddRef(lpCollection);
	VariantClear(&varRet);

	/*
	 *	count = Styles.Count
	 */
	hr = DispatchInvoke(lpCollection, DISPATCH_PROPERTYGET,
			OLESTR("Count"), &varRet, 0, NULL);
	if (!SUCCEEDED(hr)) {
		TRACE1("Styles.Count failed with 0x%1!08x!.\n", hr);
		lpCollection->lpVtbl->Release(lpCollection);
		return(hr);
	}
	lCount = V_I4(&varRet);
	VariantClear(&varRet);
	TRACE1(_T("Styles.Count = %1!d!\n"), lCount);

	for (l = lCount; 0 < l; l--) {
		if (bAbortOperation) break;

		/* Style = Styles->Item(index) */
		hr = DispatchInvoke(lpCollection, DISPATCH_METHOD,
				OLESTR("Item"), &varRet, 1, _T("i"), l);
		if (!SUCCEEDED(hr)) {
			TRACE2(_T("Styles->Item(%1!d!) failed with 0x%2!08x!.\n"), l, hr);
			continue;
		}
		lpItem = V_DISPATCH(&varRet);
		lpItem->lpVtbl->AddRef(lpItem);
		VariantClear(&varRet);

		/*
		 *	check Style.BuiltIn property and call Style->Delete() if
		 *	it is not a built-in style.
		 */
		hr = DispatchInvoke(lpItem, DISPATCH_PROPERTYGET,
				OLESTR("BuiltIn"), &varRet, 0, NULL);
		if (!SUCCEEDED(hr)) {
			TRACE2(_T("Item(%1!d!).BuiltIn failed with 0x%2!08x!.\n"), l, hr);
			lpItem->lpVtbl->Release(lpItem);
			continue;
		}
		bBuiltIn = V_BOOL(&varRet);
		VariantClear(&varRet);
		if (!bBuiltIn) {
			hr = DispatchInvoke(lpItem, DISPATCH_METHOD,
					OLESTR("Delete"), &varRet, 0, NULL);
			if (SUCCEEDED(hr)) {
				VariantClear(&varRet);
			} else {
				TRACE1(_T("Item(%1!d!)->Delete() failed with 0x%1!08x!.\n"),
						hr);
			}
		}
		lpItem->lpVtbl->Release(lpItem);
	}

#ifdef _DEBUG
	hr = lpCollection->lpVtbl->Release(lpCollection);
	PRINTF(STDERR, _T("debug: Styles->Release() retunred %1!d!.\n"), hr);
#else	/* !_DEBUG */
	lpCollection->lpVtbl->Release(lpCollection);
#endif	/* _DEBUG */
	if (bAbortOperation) return(E_ABORT);
	return(S_OK);
}


/*------------------------------------------------------------------------
	resetting to Normal style
------------------------------------------------------------------------*/

HRESULT WordNormalStyle(LPDISPATCH lpDoc, LPCMDPARAM lpParam)
{
	HRESULT hr;
	VARIANT varRet;
	LPDISPATCH lpRange;

	TRACE0(_T("Trying to reset text styles.\n"));
	VariantInit(&varRet);

	/*
	 *	obtain Document.Range object.
	 */
	hr = DispatchInvoke(lpDoc, DISPATCH_METHOD,
			OLESTR("Range"), &varRet, 0, NULL);
	if (!SUCCEEDED(hr)) {
		TRACE1(_T("Document.Range() failed with 0x%1!08x!.\n"), hr);
		return(hr);
	}
	lpRange = V_DISPATCH(&varRet);
	lpRange->lpVtbl->AddRef(lpRange);
	VariantClear(&varRet);

	/* Range.Style = Word::wdStyleNormal */
	hr = DispatchInvoke(lpRange, DISPATCH_PROPERTYPUT,
			OLESTR("Style"), &varRet, 1, _T("i"), -1);
	if (!SUCCEEDED(hr)) {
		TRACE1(_T("Setting Range.Style to wdStyleNormal ")
				_T("failed with 0x%1!08x!.\n"), hr);
	}

	/* Range.Style = Word::wdStyleDefaultParagraphFont */
	hr = DispatchInvoke(lpRange, DISPATCH_PROPERTYPUT,
			OLESTR("Style"), &varRet, 1, _T("i"), -66);
	if (!SUCCEEDED(hr)) {
		TRACE1(_T("Setting Range.Style to wdStyleDefaultParagraphFont ")
				_T("failed with 0x%1!08x!.\n"), hr);
	}

#ifdef _DEBUG
	hr = lpRange->lpVtbl->Release(lpRange);
	PRINTF(STDERR, _T("debug: lpRange->Release() returned %1!d!.\n"), hr);
#else	/* !_DEBUG */
	lpRange->lpVtbl->Release(lpRange);
#endif	/* _DEBUG */
	return(S_OK);
}


/*------------------------------------------------------------------------
	calling SaveAs method
------------------------------------------------------------------------*/

BOOL WordSaveAs(LPCSTR lpszFrom, LPCMDPARAM lpParam)
{
	HRESULT hr;
	VARIANT varRet;
	LPDISPATCH lpDoc;
	TCHAR szPath[_MAX_PATH + 1];

	/*
	 *	make the output filename.
	 */
	if (lpParam->dwFlags & FLAG_STDOUT)
		GetTempFileName(lpParam->szOutDir, _T("doc"), 0, szPath);
	else if (!MakeOutputPath(lpszFrom, szPath, TCCH(szPath) - 1, lpParam))
		return(FALSE);

	/*
	 *	run the Word application if not yet.
	 */
	if (!(lpParam->lpApp)) {
		WordInitialize(lpParam);
		if (bAbortOperation) return(FALSE);
	}

	/*
	 *	open the input file in read-only mode.
	 */
	hr = DispatchInvoke(lpParam->lpDocs, DISPATCH_METHOD, OLESTR("Open"),
			&varRet, 4, _T("sbbb"), lpszFrom, FALSE, TRUE, FALSE);
	if (!SUCCEEDED(hr)) {
		Gripe(MSG_W_WORD_OPEN, hr, lpszFrom);
		return(FALSE);
	}
	lpDoc = V_DISPATCH(&varRet);
	lpDoc->lpVtbl->AddRef(lpDoc);
	VariantClear(&varRet);

	/*
	 *	remove ornaments according to the given command-line options.
	 *	ignore errors during this process.
	 */
	if (!bAbortOperation && (lpParam->dwFlags & FLAG_REMOVE_COMMENT))
		WordRemoveComment(lpDoc, lpParam);
	if (!bAbortOperation && (lpParam->dwFlags & FLAG_REMOVE_STYLE)) {
		WordRemoveStyle(lpDoc, lpParam);
		if (!bAbortOperation) WordNormalStyle(lpDoc, lpParam);
	}

	/*
	 *	save with the new filename.
	 */
	if (bAbortOperation) goto CLOSE;
	if (!(lpParam->dwFlags & (FLAG_STDOUT | FLAG_QUIET)))
		PRINTF(STDOUT, _T("%1 --> %2%n"), lpszFrom, szPath);
	if (lpParam->dwFlags & FLAG_DEBUG) goto CLOSE;

	lstrcat(szPath, _T("."));
	hr = DispatchInvoke(lpDoc, DISPATCH_METHOD,
			OLESTR("SaveAs"), NULL, 5, _T("sibSb"),
			szPath, lpParam->nSaveFmt, FALSE, NULL, FALSE);

CLOSE:
	/*
	 *	close the input file.
	 */
	DispatchInvoke(lpDoc, DISPATCH_METHOD,
			OLESTR("Close"), NULL, 2, _T("i"), 0);
	lpDoc->lpVtbl->Release(lpDoc);

	if (!SUCCEEDED(hr)) {
		Gripe(MSG_W_WORD_SAVEAS, hr);
		return(FALSE);
	}
	if (!(lpParam->dwFlags & FLAG_STDOUT) || bAbortOperation)
		return(!bAbortOperation);

	/*
	 *	write the result on the standard output. this must be done after
	 *	the output file is closed, or access to the file is denined.
	 */
	hr = S_OK;
	if (ConPrintFile(STDOUT, szPath)) PRINT(STDOUT, _T("\n"));
	else if (!bAbortOperation) hr = GetLastError();
	DeleteFile(szPath);

	if (WDFMT_HTML == lpParam->nSaveFmt && !(lpParam->dwFlags & FLAG_DEBUG)) {
		/*
		 *	if the output format is HTML, a directory that contains
		 *	the image files may have been created.
		 */
		LPTSTR lpstr;

		szPath[lstrlen(szPath) - 1] = 0;
		lpstr = _tcsrchr(szPath, _TCHAR('.'));
		if (!lpstr) lpstr = szPath + lstrlen(szPath);
		lstrcpy(lpstr, _T(".files"));
		DeleteDirectory(szPath);
	}

	if (hr) {
		Gripe(MSG_W_PRINT_OUTFILE, hr);
		return(FALSE);
	}
	return(!bAbortOperation);
}


/*------------------------------------------------------------------------
	conversion start point
------------------------------------------------------------------------*/

int ConvertFile(LPCSTR lpszPattern, LPCMDPARAM lpParam)
{
	LPSTR lpszFileFrom;
	LPCSTR lpszFileOrig;
	DWORD dwLen;
	HANDLE hFind;
	WIN32_FIND_DATA wfd;
	int nFound = 0;
	char szPath[_MAX_PATH];

	if (bAbortOperation) return(0);

	/*
	 *	fix the input path.
	 */
	dwLen = FixPath(lpszPattern, szPath, TCCH(szPath), &lpszFileFrom,
			(BOOL)!(lpParam->dwFlags & FLAG_NO_CYGWIN_PATH));
	if (!dwLen) Gripe(MSG_E_UNEXPECTED, GetLastError());
	dwLen = (DWORD)(lpszFileFrom - szPath);

	TRACE1(_T("Searching \"%1\"...\n"),  szPath);

	/*
	 *	find the first file that matches the given pattern.
	 */
	hFind = FindFirstFile(szPath, &wfd);
	if (INVALID_HANDLE_VALUE == hFind) goto RECURSIVE;

	/*
	 *	process all matched files.
	 */
	do {
		if (wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) continue;
		if (dwLen + lstrlen(wfd.cFileName) < TCCH(szPath)) {
			lstrcpy(lpszFileFrom, wfd.cFileName);
			WordSaveAs(szPath, lpParam);
			nFound++;
		} else {
			*lpszFileFrom = 0;
			Gripe(MSG_W_TOO_LONG_FILENAME, 0, szPath, wfd.cFileName, "");
		}
	} while (!bAbortOperation && FindNextFile(hFind, &wfd));

	/*
	 *	close the search handle.
	 */
	FindClose(hFind);

RECURSIVE:
	if (bAbortOperation || !(lpParam->dwFlags & FLAG_RECURSIVE))
		return(nFound);

	/*
	 *	find the same filename from subdirectories recursively.
	 */
	lpszFileOrig = lpszPattern + lstrlen(lpszPattern);
	while (lpszPattern < lpszFileOrig) {
		lpszFileOrig = CharPrev(lpszPattern, lpszFileOrig);
		if (IsDBCSLeadByte(*lpszFileOrig)) continue;
		if ((TCHAR)'\\' == *lpszFileOrig
				|| (TCHAR)'/' == *lpszFileOrig
				|| (TCHAR)':' == *lpszFileOrig) {
			lpszFileOrig++;
			break;
		}
	}

	dwLen = (DWORD)(lpszFileFrom - szPath) + lstrlen(lpszFileOrig) + 1;
	lstrcpy(lpszFileFrom, _T("*.*"));
	hFind = FindFirstFile(szPath, &wfd);
	if (INVALID_HANDLE_VALUE == hFind) return(nFound);

	do {
		if ((wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
				&& lstrcmp(_T("."), wfd.cFileName)
				&& lstrcmp(_T(".."), wfd.cFileName)) {
			if (dwLen + lstrlen(wfd.cFileName) < TCCH(szPath)) {
				wsprintf(lpszFileFrom, _T("%s\\%s"),
						wfd.cFileName, lpszFileOrig);
				nFound += ConvertFile(szPath, lpParam);
			} else {
				*lpszFileFrom = 0;
				Gripe(MSG_W_TOO_LONG_FILENAME, 0,
						szPath, wfd.cFileName, lpszFileOrig);
			}
		}
	} while (!bAbortOperation && FindNextFile(hFind, &wfd));

	/*
	 *	close the search handle.
	 */
	FindClose(hFind);
	return(nFound);
}
