/* $Id: dispatch.c 1.5.0.1 2000/12/22 00:26:50 satomii Exp $ */
/*----------------------------------------------------------------------*/
/*		dispatch.c														*/
/*			designed		by satomii		00/12/17					*/
/*			copyright (c) 1999-2000, satomii. all rights reserved.		*/
/*----------------------------------------------------------------------*/
/*
 *	this program is free software; you can redistribute and/or modify it
 *	under the terms of the GNU General Public License either version 2 or
 *	later, as published by the Free Software Foundation.
 *
 *	this program is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *	General Public License for more details.
 */
#include <windows.h>

#define _TCHAR(c)			((TCHAR)(c))


LPWSTR ToWideChar(LPCSTR lpszText)
{
	DWORD len;
	LPWSTR lpwstr;

	len = MultiByteToWideChar(CP_OEMCP, 0, lpszText, -1, NULL, 0);
	lpwstr = (LPWSTR)LocalAlloc(LPTR, sizeof(WCHAR) * len);
	if (lpwstr)
		MultiByteToWideChar(CP_OEMCP, 0, lpszText, -1, lpwstr, len);
	return(lpwstr);
}

BSTR ToBSTR(LPCTSTR lpszText)
{
	BSTR bstr = NULL;

	LPWSTR lpwstr;
	lpwstr = ToWideChar(lpszText);
	if (lpwstr) {
		bstr = SysAllocString(lpwstr);
		LocalFree(lpwstr);
	}
	return(bstr);
}

static LPVARIANTARG GetVariantArgs(UINT argc,
		LPCTSTR argv, va_list *const pva)
{
	LPVARIANTARG lpArgs, lpVar;
	LPVOID lpvParam;

	lpArgs = (LPVARIANTARG)LocalAlloc(LPTR, sizeof(VARIANTARG) * argc);
	if (!lpArgs) return(NULL);
	lpVar = lpArgs + argc - 1;

	while (0 < argc--) {
		VariantInit(lpVar);
		switch (*argv++) {
			case _TCHAR('i'):	/* 32-bit integer. */
				V_VT(lpVar) = VT_I4;
				V_I4(lpVar) = va_arg(*pva, long);
				break;
			case _TCHAR('b'):	/* bool. */
				V_VT(lpVar) = VT_BOOL;
				V_BOOL(lpVar) = va_arg(*pva, VARIANT_BOOL);
				break;
			case _TCHAR('s'):	/* ANSI string. */
				V_VT(lpVar) = VT_BSTR;
				lpvParam = va_arg(*pva, LPSTR);
				if (lpvParam) V_BSTR(lpVar) = ToBSTR((LPSTR)lpvParam);
				else V_BSTR(lpVar) = NULL;
				break;
			case _TCHAR('S'):	/* OLE (unicode) string. */
				V_VT(lpVar) = VT_BSTR;
				lpvParam = va_arg(*pva, OLECHAR *);
				if (lpvParam)
					V_BSTR(lpVar) = SysAllocString((OLECHAR *)lpvParam);
				else V_BSTR(lpVar) = NULL;
				break;
			case _TCHAR('e'):	/* empty. */
			default:			/* not used, leave as empty. */
				break;
		}
		lpVar--;
	}

	return(lpArgs);
}

HRESULT DispatchInvoke(LPDISPATCH lpObj, WORD wFlags,
		LPOLESTR lpwszName, LPVARIANT lpvRet, UINT argc, LPCTSTR argv, ...)
{
	static DISPID dispidNamed = DISPID_PROPERTYPUT;
	HRESULT hr;
	DISPID dispid;
	DISPPARAMS dispParams;
	UINT i;

	/* get the DISPID of the property/method name to invoke. */
	hr = lpObj->lpVtbl->GetIDsOfNames(lpObj, &IID_NULL,
			&lpwszName, 1, LOCALE_SYSTEM_DEFAULT, &dispid);
	if (!SUCCEEDED(hr)) return(hr);

	/* fill the DISPPARAM structure. */
	ZeroMemory(&dispParams, sizeof(dispParams));
	if (wFlags & DISPATCH_PROPERTYPUT) {
		dispParams.rgdispidNamedArgs = &dispidNamed;
		dispParams.cNamedArgs = 1;
	}

	if (0 < argc) {
		va_list va;

		va_start(va, argv);
		dispParams.rgvarg = GetVariantArgs(argc, argv, &va);
		va_end(va);
		if (!(dispParams.rgvarg)) return(E_OUTOFMEMORY);
		dispParams.cArgs = argc;
	}

	/* call the IDispatch::Invoke method. */
	if (lpvRet) VariantInit(lpvRet);
	hr = lpObj->lpVtbl->Invoke(lpObj, dispid, &IID_NULL,
			LOCALE_SYSTEM_DEFAULT, wFlags, &dispParams, lpvRet, NULL, NULL);

	/* clean up. */
	for (i = 0; i < argc; i++) VariantClear(dispParams.rgvarg + i);
	if (dispParams.rgvarg) LocalFree(dispParams.rgvarg);
	return(hr);
}
