/* $Id: doc2txt.h 1.5 2000/12/21 11:56:51 satomii Exp $ */
/*----------------------------------------------------------------------*/
/*		doc2txt.h														*/
/*			designed		by satomii		00/12/15					*/
/*			copyright (c) 1999-2000, satomii. all rights reserved.		*/
/*----------------------------------------------------------------------*/
/*
 *	this program is free software; you can redistribute and/or modify it
 *	under the terms of the GNU General Public License either version 2 or
 *	later, as published by the Free Software Foundation.
 *
 *	this program is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *	General Public License for more details.
 */
#ifndef __DOC2TXT_H__
#define __DOC2TXT_H__

#include <tchar.h>

#if defined UNICODE || defined _UNICODE
#define TCCH(buf)					(sizeof(buf) / sizeof(TCHAR))
#define TCB(num)					(num * sizeof(TCHAR))
#else	/* !UNICODE && !_UNICODE */
#define TCCH(buf)					sizeof(buf)
#define TCB(num)					(num)
#endif	/* UNICODE || _UNICODE */
#define _TCHAR(c)					((TCHAR)(c))


/*------------------------------------------------------------------------
	type definitions
------------------------------------------------------------------------*/

typedef enum
{
	WDFMT_INVALID = -1,
    WDFMT_DOCUMENT = 0,
    WDFMT_TEMPLATE = 1,
    WDFMT_TEXT = 2,
    WDFMT_TEXTLINEBREAKS = 3,
    WDFMT_DOSTEXT = 4,
    WDFMT_DOSTEXTLINEBREAKS = 5,
    WDFMT_RTF = 6,
    WDFMT_UNICODETEXT = 7,
    WDFMT_ENCODEDTEXT = 7,
    WDFMT_HTML = 8,
	WDFMT_LAST = 9
} WDSAVEFMT;

#pragma pack(1)

typedef struct _tagCMDPARAM {
	DWORD dwFlags;
	DWORD dwOutDirLen;
	DWORD dwOutExtLen;
	WDSAVEFMT nSaveFmt;
	LPCTSTR lpszProgID;
	LPDISPATCH lpApp;
	LPDISPATCH lpDocs;
	TCHAR szOutDir[_MAX_DIR];
	TCHAR szOutExt[40];
} CMDPARAM, FAR *LPCMDPARAM;

#pragma pack()


/*------------------------------------------------------------------------
	macro definitions
------------------------------------------------------------------------*/

/* console handle indices. */
#define STDIN		STD_INPUT_HANDLE
#define STDOUT		STD_OUTPUT_HANDLE
#define STDERR		STD_ERROR_HANDLE

/* option flags for CMDPARAM.dwFlags. */
#define FLAG_STDOUT					0x00000001
#define FLAG_ORIGPATH				0x00000002
#define FLAG_APPEND_EXTENSION		0x00000004
#define FLAG_RECURSIVE				0x00000008
#define FLAG_NO_CYGWIN_PATH			0x00000010
#define FLAG_OVERWRITE				0x00000100
#define FLAG_SKIP_EXISTING			0x00000200
#define FLAG_EXIT_ON_WARNING		0x00001000
#define FLAG_QUIET					0x00002000
#define FLAG_DEBUG					0x00004000
#define FLAG_REMOVE_STYLE			0x00010000
#define FLAG_REMOVE_COMMENT			0x00020000


/*------------------------------------------------------------------------
	debugging macros
------------------------------------------------------------------------*/

#define TRACE0(s)			if (DEBUG) PRINT(STDOUT, _T("DEBUG: ") s)
#define TRACE1(s, a1)		if (DEBUG) PRINTF(STDOUT, _T("DEBUG: ") s, a1)
#define TRACE2(s, a1, a2)	if (DEBUG) PRINTF(STDOUT, _T("DEBUG: ") s, a1, a2)


/*------------------------------------------------------------------------
	functions
------------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C" {
#endif	/* __cplusplus */

/* console.c */
DWORD ConPrint(DWORD, LPCTSTR);
DWORD ConPrintFormat(DWORD, LPCTSTR, ...);
DWORD ConPrintFile(DWORD, LPCTSTR);

#ifndef PRINT
#define PRINT		ConPrint
#endif	/* !PRINT */
#ifndef PRINTF
#define PRINTF		ConPrintFormat
#endif	/* !PRINTF */

/* conv.c */
int ConvertFile(LPCSTR, struct _tagCMDPARAM FAR *);

/* dispatch.c */
LPWSTR ToWideChar(LPCSTR);
BSTR ToBSTR(LPCTSTR);
HRESULT DispatchInvoke(LPDISPATCH, WORD, LPOLESTR, LPVARIANT, UINT, LPCTSTR, ...);

/* main.c */
void Gripe(DWORD, DWORD, ...);
void Trace(LPCTSTR, ...);
void CleanupDocTo(void);

extern int bAbortOperation;

/* option.c */
int GetOption(int, LPCTSTR *, LPCMDPARAM);

/* path.c */
DWORD FixPath(LPCTSTR, LPTSTR, DWORD, LPTSTR *, BOOL);

#ifdef __cplusplus
}
#endif	/* __cplusplus */


#endif	/* __DOC2TXT_H__ */
