/* $Id: main.c 1.5.0.2 2000/12/22 01:59:00 satomii Exp $ */
/*----------------------------------------------------------------------*/
/*		main.c															*/
/*			designed		by satomii		00/12/15					*/
/*			copyright (c) 1999-2000, satomii. all rights reserved.		*/
/*----------------------------------------------------------------------*/
/*
 *	this program is free software; you can redistribute and/or modify it
 *	under the terms of the GNU General Public License either version 2 or
 *	later, as published by the Free Software Foundation.
 *
 *	this program is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *	General Public License for more details.
 */
#include <windows.h>
#include "doc2txt.h"
#include "msgtbl.h"

static CMDPARAM param;
int bAbortOperation = FALSE;

#define DEBUG					(param.dwFlags & FLAG_DEBUG)


/*------------------------------------------------------------------------
	message output
------------------------------------------------------------------------*/

void Gripe(DWORD dwMsgId, DWORD dwErrCode, ...)
{
	static TCHAR *ppszSeverity[] = { _T("Warning"), _T("Error") };
	LPTSTR lpstr;
	BYTE bSeverity;
	DWORD fp;

#define MESSAGE_TO_BE_PRINTED			\
		(!(param.dwFlags & FLAG_QUIET)	\
		|| (bSeverity & 0x01)			\
		|| (2 == bSeverity && (param.dwFlags & FLAG_EXIT_ON_WARNING)))

	bSeverity = (BYTE)((dwMsgId >> 30) & 0x03);
	fp = (bSeverity < 2 || DEBUG) ? STDOUT : STDERR;

	if (MESSAGE_TO_BE_PRINTED) {
		va_list va;

		va_start(va, dwErrCode);
		FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
				FORMAT_MESSAGE_FROM_HMODULE,
				GetModuleHandle(NULL), dwMsgId, 0, (LPTSTR)&lpstr, 0, &va);
		va_end(va);
		if (lpstr) {
			if (2 <= bSeverity)
				PRINTF(fp, _T("%1: "), ppszSeverity[bSeverity - 2]);
			PRINTF(fp, lpstr);
			LocalFree(lpstr);
		} else if (2 <= bSeverity) {
			PRINTF(fp, _T("%1 0x%2!08x! has occurred.\n"),
					ppszSeverity[bSeverity - 2], dwMsgId);
		}
	}

	if (bSeverity < 2) return;	/* informatinal message. */

	if (dwErrCode && MESSAGE_TO_BE_PRINTED) {
		FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
				FORMAT_MESSAGE_FROM_SYSTEM |
				FORMAT_MESSAGE_IGNORE_INSERTS,
				NULL, dwErrCode, 0, (LPTSTR)&lpstr, 0, NULL);
		PRINTF(fp, HIWORD(dwErrCode) ? _T("[%1!08x!] ") : _T("[%1!d!] "),
				dwErrCode);
		if (lpstr) {
			PRINTF(fp, lpstr);
			LocalFree(lpstr);
		} else {
			PRINTF(fp, _T("(Unknown error)\n"));
		}
	}

	if (3 == bSeverity || (param.dwFlags & FLAG_EXIT_ON_WARNING)) {
		CleanupDocTo();
		ExitProcess(3);
	}
}


/*------------------------------------------------------------------------
	clean-up routine
------------------------------------------------------------------------*/

void CleanupDocTo(void)
{
#ifdef _DEBUG
	PRINT(STDERR, _T("debug: CleanupDocTo() called.\n"));
#endif	/* _DEBUG */

	if (param.lpDocs) {
		TRACE0(_T("Releasing Documents collection object.\n"));
		param.lpDocs->lpVtbl->Release(param.lpDocs);
		param.lpDocs = NULL;
	}

	if (param.lpApp) {
		HRESULT hr;

		hr = DispatchInvoke(param.lpApp, DISPATCH_METHOD,
				OLESTR("Quit"), NULL, 1, _T("i"), 0);
		if (!SUCCEEDED(hr)) {
			Gripe(MSG_W_WORD_QUIT, hr);
			DispatchInvoke(param.lpApp, DISPATCH_PROPERTYPUT,
					OLESTR("Visible"), NULL, 1, _T("b"), TRUE);
		}
		TRACE0(_T("Word successfully closed.\n"));
		TRACE0(_T("Releasing Application object.\n"));
		param.lpApp->lpVtbl->Release(param.lpApp);
		param.lpApp = NULL;
		OleUninitialize();
	}
}

BOOL WINAPI CleanupHandler(DWORD dwCtrlType)
{
	HANDLE hStdIn;
	DWORD dw;

	UNREFERENCED_PARAMETER(dwCtrlType);
#ifdef _DEBUG
	PRINT(STDERR, _T("\ndebug: CleanupHandler called, cleaning up.\n"));
#endif	/* _DEBUG */

	hStdIn = GetStdHandle(STDIN);
	if (GetConsoleMode(hStdIn, &dw)) {
		dw |= ENABLE_PROCESSED_OUTPUT;
		SetConsoleMode(GetStdHandle(STDIN), dw);
	}
	bAbortOperation = TRUE;
	return(TRUE);
}


/*------------------------------------------------------------------------
	the main routine
------------------------------------------------------------------------*/

int main(int argc, char **argv)
{
	int nParam, nErr = 0, i;
	extern const TCHAR *ppszFmtName[];

	/*
	 *	parse the command line.
	 */
	nParam = GetOption(argc, (LPCTSTR *)argv, &param);

#ifdef _DEBUG
	param.dwFlags |= FLAG_DEBUG;
#endif	/* _DEBUG */

	if (DEBUG) {
		TRACE0(_T("Working in debug mode.\n\n"));

		TRACE1(_T("%1!d! arguments to be processed"), argc - nParam);
		if (param.dwFlags & FLAG_RECURSIVE) PRINT(STDOUT, _T(" recursively"));
		PRINT(STDOUT, _T(".\n"));

		for (i = nParam; i < argc; i++)
			TRACE2(_T("  argv[%1!d!] = %2%n"), i, argv[i]);

		TRACE2(_T("Output format: %1!d! (%2)%n"),
				param.nSaveFmt, ppszFmtName[param.nSaveFmt]);
		TRACE1(_T("Output directory: %1%n"),
				param.szOutDir[0] ? param.szOutDir : _T("(input directory)"));
		TRACE2(_T("Output extension: %1%2"), param.szOutExt,
				(param.dwFlags & FLAG_APPEND_EXTENSION) ?
				_T(" (append)") : _T(""));

		PRINT(STDOUT, _T("\n\n"));
	}

	SetConsoleCtrlHandler(CleanupHandler, TRUE);

	/*
	 *	dispatch input files.
	 */
	while (nParam < argc) {
		if (bAbortOperation) break;
		if (!ConvertFile(*(argv + nParam), &param)) {
			Gripe(MSG_W_FILE_NOT_FOUND, GetLastError(), *(argv + nParam));
			nErr++;
		}
		nParam++;
	}

	/*
	 *	clean up.
	 */
	CleanupDocTo();
	if (bAbortOperation) {
		Gripe(MSG_I_ABORTED, 0);
		return(3);
	}
	return(nErr ? 1 : 0);
}
