/* $Id: option.c 1.5.0.1 2000/12/22 00:26:50 satomii Exp $ */
/*----------------------------------------------------------------------*/
/*		option.c														*/
/*			designed		by satomii		00/12/15					*/
/*			copyright (c) 1999-2000, satomii. all rights reserved.		*/
/*----------------------------------------------------------------------*/
/*
 *	a callback function for ParseCommandLine().
 *
 *	this program is free software; you can redistribute and/or modify it
 *	under the terms of the GNU General Public License either version 2 or
 *	later, as published by the Free Software Foundation.
 *
 *	this program is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *	General Public License for more details.
 */
#include <windows.h>
#include "doc2txt.h"
#include "cmdline.h"
#include "msgtbl.h"

const OPTIONFORMAT option[] = {
	{ _TCHAR('f'), _TCHAR('='),	_T("f"), _T("format")			},
	{ _TCHAR('o'), _TCHAR('='),	_T("o"), _T("outdir")			},
	{ _TCHAR('O'), 0,			_T("O"), _T("origdir")			},
	{ _TCHAR('e'), _TCHAR('='),	_T("e"), _T("change-extension")	},
	{ _TCHAR('E'), _TCHAR('='),	_T("E"), _T("append-extension")	},
	{ _TCHAR('s'), 0,			_T("s"), _T("stdout")			},
	{ _TCHAR('c'), 0,			_T("c"), _T("no-cygwin-path")	},
	{ _TCHAR('r'), 0,			_T("r"), _T("recursive")		},
	{ _TCHAR('w'), 0,			_T("w"), _T("overwrite")		},
	{ _TCHAR('k'), 0,			_T("k"), _T("skip-existing")	},
	{ _TCHAR('P'), _TCHAR('='),	_T("P"), _T("progid")			},
	{ _TCHAR('S'), 0,			_T("S"), _T("remove-style")		},
	{ _TCHAR('C'), 0,			_T("C"), _T("remove-comment")	},
	{ _TCHAR('W'), 0,			_T("W"), _T("exit-on-warning")	},
	{ _TCHAR('q'), 0,			_T("q"), _T("quiet")			},
	{ _TCHAR('D'), 0,			_T("D"), _T("debug")			},
	{ _TCHAR('v'), 0,			_T("v"), _T("version")			},
	{ _TCHAR('h'), 0,			_T("h?"), _T("help")			}
};

const TCHAR szFmtAbbrev[] = _T("wltTdDruh");

const TCHAR *ppszFmtName[WDFMT_LAST] = {
	_T("wdFormatDocument"),
	_T("wdFormatTemplate"),
	_T("wdFormatText"),
	_T("wdFormatTextLineBreaks"),
	_T("wdFormatDOSText"),
	_T("wdFormatDOSTextLineBreaks"),
	_T("wdFormatRTF"),
	_T("wdFormatUnicodeText"),
	_T("wdFormatHTML")
};

const TCHAR *ppszFmtExt[WDFMT_LAST] = {
	  /* 0 */    /* 1 */    /* 2 */    /* 3 */    /* 4 */    /* 5 */
	_T("doc"), _T("dot"), _T("txt"), _T("txt"), _T("txt"), _T("txt"),
	  /* 6 */    /* 7 */    /* 8 */
	_T("rtf"), _T("txt"), _T("html")
};


/*------------------------------------------------------------------------
	version information
------------------------------------------------------------------------*/

void Version(void)
{
	LPVOID lpvVer;
	LPSTR lpszDesc, lpszProdVer, lpszSpecBld, lpszCopyright;
	DWORD dwLen, dw;
	TCHAR szModulePath[_MAX_PATH];

	GetModuleFileName(GetModuleHandle(NULL),
			szModulePath, TCCH(szModulePath));
	dwLen = GetFileVersionInfoSize(szModulePath, &dw);
	lpvVer = (LPVOID)LocalAlloc(LPTR, dwLen);
	if (!lpvVer) Gripe(MSG_E_OUTOFMEMORY, 0);

	if (!GetFileVersionInfo(szModulePath, dw, dwLen, lpvVer)) {
		dw = GetLastError();
		LocalFree(lpvVer);
		Gripe(MSG_E_GETVERSION, dw);
	}

	VerQueryValue(lpvVer,
			_T("\\StringFileInfo\\040904b0\\FileDescription"),
			(LPVOID *)&lpszDesc, (PUINT)&dwLen);
	VerQueryValue(lpvVer,
			_T("\\StringFileInfo\\040904b0\\ProductVersion"),
			(LPVOID *)&lpszProdVer, (PUINT)&dwLen);
	VerQueryValue(lpvVer,
			_T("\\StringFileInfo\\040904b0\\SpecialBuild"),
			(LPVOID *)&lpszSpecBld, (PUINT)&dwLen);
	VerQueryValue(lpvVer,
			_T("\\StringFileInfo\\040904b0\\LegalCopyright"),
			(LPVOID *)&lpszCopyright, (PUINT)&dwLen);

	PRINT(STDOUT, _T("DOC2TXT"));
	if (lpszDesc && *lpszDesc) PRINTF(STDOUT, _T(" - %1"), lpszDesc);
	if (lpszProdVer && *lpszProdVer)
		PRINTF(STDOUT, _T(", Ver.%1"), lpszProdVer);
	if (lpszSpecBld && *lpszSpecBld)
		PRINTF(STDOUT, _T(" (%1)"), lpszSpecBld);
	if (lpszCopyright && *lpszCopyright)
		PRINTF(STDOUT, _T("%n%1"), lpszCopyright);
	PRINT(STDOUT, _T("\n"));

	LocalFree(lpvVer);
	CleanupDocTo();
	ExitProcess(0);
}


/*------------------------------------------------------------------------
	usage information
------------------------------------------------------------------------*/

void Usage(BOOL bVerbose)
{
	LPSTR lpszCmd, lpstr;
	TCHAR szModulePath[_MAX_PATH];

	GetModuleFileName(GetModuleHandle(NULL),
			szModulePath, TCCH(szModulePath));
	lpszCmd = _tcsrchr(szModulePath, _TCHAR('\\'));
	if (lpszCmd) lpszCmd++;
	else lpszCmd = szModulePath;

	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
			FORMAT_MESSAGE_FROM_HMODULE |
			FORMAT_MESSAGE_ARGUMENT_ARRAY,
			GetModuleHandle(NULL),
			bVerbose ? MSG_I_HELP : MSG_I_SYNTAX,
			0, (LPSTR)&lpstr, 0, (va_list *)&lpszCmd);
	if (lpstr) {
		PRINTF(bVerbose ? STDOUT : STDERR, lpstr);
		LocalFree(lpstr);
	} else {
		Gripe(MSG_E_UNEXPECTED, GetLastError());
	}

	CleanupDocTo();
	ExitProcess(bVerbose ? 0 : 1);
}


/*------------------------------------------------------------------------
	callback functions for ParseCommandLine()
------------------------------------------------------------------------*/

#define GRIPE_DUPLICATE()											\
		Gripe(argv2 ?												\
		MSG_E_OPTION_DUPLICATED2 : MSG_E_OPTION_DUPLICATED,		\
		0, argv1, argv2);

#define GRIPE_CONFLICT(opt) {												\
		if (argv2)															\
			Gripe(MSG_E_OPTION_CONFLICT2, 0, argv1, argv2, _TCHAR(opt));	\
		else Gripe(MSG_E_OPTION_CONFLICT, 0, argv1, _TCHAR(opt));			\
	}

#define GRIPE_INVALID_CMDLINE()										\
		Gripe(MSG_E_INVALID_CMDLINE, 0, argv1)
#define GRIPE_PARAM_REQUIRED()										\
		Gripe(MSG_E_PARAMETER_REQUIRED, 0, argv1)
#define GRIPE_PARAM_NOT_ALLOWED()									\
		Gripe(MSG_E_PARAMETER_NOT_ALLOWED, 0, argv1)
#define GRIPE_PARAM_INVALID(err)									\
		Gripe(MSG_E_PARAMETER_INVALID, err, argv1, lpszValue)

static int __stdcall ParseCommandLineCallback(int nType, WORD optch,
		LPCTSTR lpszValue, LPARAM lParam, LPCTSTR argv1, LPCTSTR argv2)
{
	LPCMDPARAM lpParam = (LPCMDPARAM)lParam;
	DWORD dwValue;
	TCHAR szBuf[_MAX_PATH];

	if (PARSECMD_NOERROR != nType) {
		if (_TCHAR('/') == *argv1 && *(argv1 + 1)) return(PARSECMD_IGNORE);
		switch (nType) {
			case PARSECMD_UNKNOWN_OPTION:
				Gripe(MSG_E_INVALID_OPTION, 0, argv1, optch);
				break;
			case PARSECMD_PARAM_REQUIRED:
				GRIPE_PARAM_REQUIRED();
				break;
			case PARSECMD_PARAM_NOT_ALLOWED:
				GRIPE_PARAM_NOT_ALLOWED();
				break;
			default:
				Gripe(MSG_E_UNEXPECTED, 0);
				break;
		}
		return(PARSECMD_ABORT);
	}

	switch (optch) {
		case _TCHAR('f'):
			if (!*lpszValue) GRIPE_PARAM_REQUIRED();
			dwValue = (DWORD)WDFMT_INVALID;

			if (!*(lpszValue + 1)) {
				if (_TCHAR('0') <= *lpszValue && *lpszValue <= _TCHAR('9')) {
					dwValue = *lpszValue - _TCHAR('0');
				} else {
					LPCTSTR lpstr = _tcschr(szFmtAbbrev, *lpszValue);
					if (lpstr) dwValue = (DWORD)(lpstr - szFmtAbbrev);
				}
			} else if (!lstrcmpi(_T("doc"), lpszValue)) {
				dwValue = WDFMT_DOCUMENT;
			} else if (!lstrcmpi(_T("dot"), lpszValue)) {
				dwValue = WDFMT_TEMPLATE;
			} else if (!lstrcmpi(_T("txt"), lpszValue)) {
				dwValue = WDFMT_TEXT;
			} else if (!lstrcmpi(_T("richtext"), lpszValue)) {
				dwValue = WDFMT_RTF;
			} else if (!lstrcmpi(_T("unicode"), lpszValue)
					|| !lstrcmpi(_T("wdFormatEncodedText"), lpszValue)
					|| !lstrcmpi(_T("EncodedText"), lpszValue)) {
				dwValue = WDFMT_UNICODETEXT;
			} else if (!lstrcmpi(_T("htm"), lpszValue)) {
				dwValue = WDFMT_HTML;
			} else {
				for (dwValue = 0; dwValue < WDFMT_LAST; dwValue++) {
					if (!lstrcmpi(ppszFmtName[dwValue], lpszValue)
							|| !lstrcmpi(ppszFmtName[dwValue] + 8, lpszValue))
						break;
				}
			}

			if ((WDSAVEFMT)dwValue == WDFMT_INVALID
					|| (WDSAVEFMT)dwValue <= WDFMT_INVALID)
				GRIPE_PARAM_INVALID(0);
			if (WDFMT_INVALID != lpParam->nSaveFmt
					&& lpParam->nSaveFmt != (WDSAVEFMT)dwValue)
				GRIPE_DUPLICATE();

			lpParam->nSaveFmt = (WDSAVEFMT)dwValue;
			break;

		case _TCHAR('o'):
			if (!*lpszValue) GRIPE_PARAM_REQUIRED();
			if (lpParam->dwFlags & FLAG_STDOUT) GRIPE_CONFLICT('s');
			if (lpParam->dwFlags & FLAG_ORIGPATH) GRIPE_CONFLICT('O');

			FixPath(lpszValue, szBuf, TCCH(lpParam->szOutDir) - 1, NULL,
					(BOOL)!(lpParam->dwFlags & FLAG_NO_CYGWIN_PATH));

			if (lpParam->szOutDir[0] && lstrcmpi(szBuf, lpParam->szOutDir))
				GRIPE_DUPLICATE();
			dwValue = GetFileAttributes(szBuf);
			if ((DWORD)-1 == dwValue
					|| !(dwValue & FILE_ATTRIBUTE_DIRECTORY))
				GRIPE_PARAM_INVALID(GetLastError());

			lstrcpy(lpParam->szOutDir, szBuf);
			break;

		case _TCHAR('O'):
			if (lpParam->dwFlags & FLAG_STDOUT) GRIPE_CONFLICT('s');
			if (lpParam->szOutDir[0]) GRIPE_CONFLICT('o');
			lpParam->dwFlags |= FLAG_ORIGPATH;
			break;

		case _TCHAR('e'):
		case _TCHAR('E'):
			if (lpParam->dwFlags & FLAG_STDOUT) GRIPE_CONFLICT('s');
			if (!*lpszValue) GRIPE_PARAM_REQUIRED();

			if (lpszValue) {
				if (_TCHAR('.') == *lpszValue) lpszValue++;
				if (TCCH(lpParam->szOutExt) <= lstrlen(lpszValue) + 1)
					GRIPE_PARAM_INVALID(ERROR_BUFFER_OVERFLOW);
			}
			if (_tcspbrk(lpszValue, _T(":\\/<>|*?")))
				GRIPE_PARAM_INVALID(ERROR_INVALID_NAME);

			lstrcpy(szBuf, _T("."));
			lstrcpy(szBuf + 1, lpszValue);

			if (lpParam->szOutExt[0]) {
				if (_TCHAR('e') == optch) {
					if (lpParam->dwFlags & FLAG_APPEND_EXTENSION)
						GRIPE_CONFLICT('E');
				} else {	/* 'E' */
					if (!(lpParam->dwFlags & FLAG_APPEND_EXTENSION))
						GRIPE_CONFLICT('e');
				}
				if (lstrcmp(szBuf, lpParam->szOutDir)) GRIPE_DUPLICATE();
			}

			lstrcpy(lpParam->szOutExt, szBuf);
			if (_TCHAR('E') == optch)
				lpParam->dwFlags |= FLAG_APPEND_EXTENSION;
			break;

		case _TCHAR('s'):
			if (lpParam->dwFlags & FLAG_ORIGPATH) GRIPE_CONFLICT('O');
			if (lpParam->szOutDir[0]) GRIPE_CONFLICT('E');

			if (lpParam->dwFlags & FLAG_APPEND_EXTENSION)
				GRIPE_CONFLICT('E');
			if (lpParam->szOutExt[0]) GRIPE_CONFLICT('e');

			if (lpParam->dwFlags & FLAG_OVERWRITE) GRIPE_CONFLICT('w');
			if (lpParam->dwFlags & FLAG_SKIP_EXISTING) GRIPE_CONFLICT('k');

			lpParam->dwFlags |= FLAG_STDOUT;
			break;

		case _TCHAR('c'):
			lpParam->dwFlags |= FLAG_NO_CYGWIN_PATH;
			break;
		case _TCHAR('r'):
			lpParam->dwFlags |= FLAG_RECURSIVE;
			break;
		case _TCHAR('w'):
			if (lpParam->dwFlags & FLAG_STDOUT) GRIPE_CONFLICT('s');
			if (lpParam->dwFlags & FLAG_SKIP_EXISTING) GRIPE_CONFLICT('k');
			lpParam->dwFlags |= FLAG_OVERWRITE;
			break;
		case _TCHAR('k'):
			if (lpParam->dwFlags & FLAG_STDOUT) GRIPE_CONFLICT('s');
			if (lpParam->dwFlags & FLAG_OVERWRITE) GRIPE_CONFLICT('w');
			lpParam->dwFlags |= FLAG_SKIP_EXISTING;
			break;
		case _TCHAR('P'):
			if (lpParam->lpszProgID
					&& lstrcmpi(lpszValue, lpParam->lpszProgID))
				GRIPE_DUPLICATE();
			lpParam->lpszProgID = lpszValue;
			break;
		case _TCHAR('S'):
			lpParam->dwFlags |= FLAG_REMOVE_STYLE;
			break;
		case _TCHAR('C'):
			lpParam->dwFlags |= FLAG_REMOVE_COMMENT;
			break;
		case _TCHAR('W'):
			lpParam->dwFlags |= FLAG_EXIT_ON_WARNING;
			break;
		case _TCHAR('q'):
			lpParam->dwFlags |= FLAG_QUIET;
			break;
		case _TCHAR('D'):
			lpParam->dwFlags |= FLAG_DEBUG;
			break;
		case _TCHAR('v'):
			Version();
			break;
		case _TCHAR('h'):
			Usage(TRUE);
			break;
		default:
			break;
	}

	return(PARSECMD_ACCEPT);
}


/*------------------------------------------------------------------------
	version information
------------------------------------------------------------------------*/

int GetOption(int argc, LPCTSTR *argv, LPCMDPARAM lpParam)
{
	int nParam;

	if (1 == argc) Usage(FALSE);

	ZeroMemory(lpParam, sizeof(CMDPARAM));
	lpParam->nSaveFmt = WDFMT_INVALID;
	nParam = ParseCommandLine(argc - 1, argv + 1,
			sizeof(option) / sizeof(option[0]), option,
			ParseCommandLineCallback, (LPARAM)lpParam);
	if (argc - 1 <= nParam) Gripe(MSG_E_NO_FILENAME, 0);

	/*
	 *	ensure unspecified option parameters.
	 */
	if (WDFMT_INVALID == lpParam->nSaveFmt) lpParam->nSaveFmt = WDFMT_TEXT;

	if (lpParam->dwFlags & FLAG_STDOUT)
		GetTempPath(TCCH(lpParam->szOutDir) - 1, lpParam->szOutDir);
	else if (!(lpParam->dwFlags & FLAG_ORIGPATH)
			&& !(lpParam->szOutDir[0]))
		GetCurrentDirectory(TCCH(lpParam->szOutDir) - 1, lpParam->szOutDir);

	if (!(lpParam->dwFlags & FLAG_ORIGPATH)) {
		lpParam->dwOutDirLen = lstrlen(lpParam->szOutDir);
		if (_TCHAR('\\') != lpParam->szOutDir[lpParam->dwOutDirLen - 1]) {
			lstrcpy(lpParam->szOutDir + lpParam->dwOutDirLen, _T("\\"));
			(lpParam->dwOutDirLen)++;
		}
	}

	if (!(lpParam->szOutExt[0])) {
		lpParam->szOutExt[0] = _TCHAR('.');
		lstrcpy(lpParam->szOutExt + 1, ppszFmtExt[lpParam->nSaveFmt]);
	}
	lpParam->dwOutExtLen = lstrlen(lpParam->szOutExt);

	return(nParam + 1);
}
