/* $Id: path.c 1.5 2000/12/21 11:56:11 satomii Exp $ */
/*----------------------------------------------------------------------*/
/*		path.c															*/
/*			designed		by satomii		00/12/15					*/
/*			copyright (c) 1999-2000, satomii. all rights reserved.		*/
/*----------------------------------------------------------------------*/
/*
 *	this program is free software; you can redistribute and/or modify it
 *	under the terms of the GNU General Public License either version 2 or
 *	later, as published by the Free Software Foundation.
 *
 *	this program is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *	General Public License for more details.
 */
#include <windows.h>
#include <tchar.h>

#if defined UNICODE || defined _UNICODE
#define TSIZE(buf)			(sizeof(buf) / sizeof(TCHAR))
#else	/* !UNICODE && !_UNICODE */
#define TSIZE(buf)			sizeof(buf)
#endif	/* UNICODE || _UNICODE */
#define _TCHAR(c)			((TCHAR)(c))

DWORD FixPath(LPCTSTR, LPTSTR, DWORD, LPTSTR *, BOOL);


/*
 *	DWORD PosixlyToWin32(LPCTSTR, LPTSTR, DWORD);
 *	converts the posix (unix) like path into Win32 style and returns the
 *	length of the converted string. lpszFrom and lpszTo cannot be the same
 *	when lpszFrom contains references to environment variables.
 */
static DWORD PosixlyToWin32(LPCTSTR lpszFrom, LPTSTR lpszTo, DWORD cchTo)
{
	DWORD dwLen = 0;

	switch (*lpszFrom) {
		case _TCHAR('~'):
			if (cchTo < 7) {
				SetLastError(ERROR_BUFFER_OVERFLOW);
				return((DWORD)-1);
			}
			lstrcpy(lpszTo, _T("%HOME%"));
			lpszTo += 6;
			dwLen = 6;
			lpszFrom++;
			break;
		case _TCHAR('\\'):
			if (++dwLen < cchTo) *lpszTo++ = _TCHAR('\\');
			lpszFrom++;
			break;
		case _TCHAR('/'):
			if (_TCHAR('/') == *(lpszFrom + 1)
					&& *(lpszFrom + 2)
					&& _TCHAR('/') == *(lpszFrom + 3)) {
					/* drive name must be included, such as "//c/temp". */
				if (++dwLen < cchTo) *lpszTo++ = *(lpszFrom + 2);
				if (++dwLen < cchTo) *lpszTo++ = _TCHAR(':');
				lpszFrom += 3;
			} else {
				if (++dwLen < cchTo) *lpszTo++ = _TCHAR('\\');
				lpszFrom++;
			}
			break;
		default:
			break;
	}

	while (*lpszFrom) {
		if (IsDBCSLeadByte(*lpszFrom)) {
			dwLen += 2;
			if (dwLen < cchTo) {
				*lpszTo++ = *lpszFrom;
				*lpszTo++ = *(lpszFrom + 1);
			}
			lpszFrom += 2;

		} else if (_TCHAR('/') == *lpszFrom) {
			if (++dwLen < cchTo) *lpszTo++ = _TCHAR('\\');
			while (_TCHAR('/') == *lpszFrom) lpszFrom++;

		} else if (_TCHAR('\\') == *lpszFrom) {
			/*
			 *	should be an escape sequence, but leave as it is unless
			 *	the next character is also '\\'.
			 */
			if (++dwLen < cchTo) *lpszTo++ = *lpszFrom;
			if (_TCHAR('\\') == *(++lpszFrom)) lpszFrom++;

		} else if (_TCHAR('$') == *lpszFrom) {
			/*
			 *	convert unix-like reference to an environment variable
			 *	into Win32 style.
			 */
			if (++dwLen < cchTo) *lpszTo++ = _TCHAR('%');
			lpszFrom++;

			if (_TCHAR('{') == *lpszFrom) {
				lpszFrom++;
				while (*lpszFrom) {
					if (IsDBCSLeadByte(*lpszFrom)) {
						dwLen += 2;
						if (dwLen < cchTo) {
							*lpszTo++ = *lpszFrom;
							*lpszTo++ = *(lpszFrom + 1);
						}
						lpszFrom += 2;
					} else if (_TCHAR('}') == *lpszFrom) {
						lpszFrom++;
						break;
					} else {
						if (++dwLen < cchTo) *lpszTo++ = *lpszFrom;
						lpszFrom++;
					}
				}
			} else {
				while (*lpszFrom) {
					if (IsDBCSLeadByte(*lpszFrom)) {
						dwLen += 2;
						if (dwLen < cchTo) {
							*lpszTo++ = *lpszFrom;
							*lpszTo++ = *(lpszFrom + 1);
						}
						lpszFrom += 2;
					} else if (_TCHAR('\\') == *lpszFrom
							|| _TCHAR('/') == *lpszFrom) {
						break;
					} else {
						if (++dwLen < cchTo) *lpszTo++ = *lpszFrom;
						lpszFrom++;
					}
				}
			}
			if (++dwLen < cchTo) *lpszTo++ = _TCHAR('%');

		} else {
			if (++dwLen < cchTo) *lpszTo++ = *lpszFrom;
			lpszFrom++;
		}
	}

	if (lpszTo) *lpszTo = 0;
	if (cchTo <= dwLen) SetLastError(ERROR_BUFFER_OVERFLOW);
	return(dwLen);
}

/*
 *	DWORD CygwinToWin32(LPCTSTR, LPTSTR, DWORD);
 *	converts the Cygwin 1.1 mounted path to Win32 path. This function does
 *	not validate the resulting path. The relative path is remained to be
 *	relative. returns the length of the converted path.
 */
static DWORD CygwinToWin32(LPCTSTR lpszFrom, LPTSTR lpszTo, DWORD cchTo)
{
	LPTSTR lpszPath, lpstr;
	HKEY hkeyMount, hkeyDir = NULL;
	DWORD dwLen, dwErr;
	TCHAR szPrefix[16];

	dwErr = RegOpenKeyEx(HKEY_CURRENT_USER,
			_T("Software\\Cygnus Solutions\\Cygwin\\mounts v2"),
			0, KEY_READ, &hkeyMount);
	if (ERROR_SUCCESS != dwErr) {
		SetLastError(dwErr);
		return(0);
	}

	/*
	 *	check the Cygwin drive prefix.
	 */
	dwLen = sizeof(szPrefix) - 1;
	if (ERROR_SUCCESS != RegQueryValueEx(hkeyMount,
			_T("cygdrive prefix"), NULL, NULL, (LPBYTE)szPrefix, &dwLen))
		goto READMOUNT;

	if (_TCHAR('/') != szPrefix[dwLen - 1]) lstrcat(szPrefix, _T("/"));
	else dwLen--;
	if (_tcsncmp(lpszFrom, szPrefix, dwLen)
			|| _TCHAR('/') != *(lpszFrom + dwLen + 1))
		goto READMOUNT;

	RegCloseKey(hkeyMount);
	lpszFrom += dwLen;

	if (3 <= cchTo) {
		*lpszTo++ = *lpszFrom;
		*lpszTo++ = _TCHAR(':');
		cchTo -= 2;
	} else {
		lpszTo = NULL;
		cchTo = 0;
	}

	lpszFrom++;
	while (_TCHAR('/') == *lpszFrom && _TCHAR('/') == *(lpszFrom + 1))
		lpszFrom++;

	return(PosixlyToWin32(lpszFrom, lpszTo, cchTo) + 2);

READMOUNT:
	/*
	 *	read the mount registry to determine whether the corresponding
	 *	path is defined.
	 */
	dwLen = lstrlen(lpszFrom);
	lpszPath = (LPTSTR)LocalAlloc(LPTR, dwLen + 1);
	if (!lpszPath) return(0);

	lstrcpy(lpszPath, lpszFrom);
	lpstr = lpszPath + dwLen;
	if (_TCHAR('/') == *(lpstr - 1)) lpstr--;

	while (lpszPath + 1 < lpstr) {
		*lpstr = 0;
		if (ERROR_SUCCESS == RegOpenKeyEx(hkeyMount,
				lpszPath, 0, KEY_READ, &hkeyDir))
			break;
		lpstr = _tcsrchr(lpszPath, _TCHAR('/'));
	}

	RegCloseKey(hkeyMount);
	if (!lpstr) {
		LocalFree(lpszPath);
		SetLastError(ERROR_PATH_NOT_FOUND);
		return(0);
	}

	dwLen = (DWORD)(lpstr - lpszPath);
	LocalFree(lpszPath);
	lpszFrom += dwLen;

	dwLen = cchTo;
	dwErr = RegQueryValueEx(hkeyDir, _T("native"), NULL, NULL,
			(LPBYTE)lpszTo, &dwLen);
	RegCloseKey(hkeyDir);
	if (ERROR_SUCCESS != dwErr) {
		SetLastError(dwErr);
		return(0);
	}

	while (_TCHAR('/') == *lpszFrom && _TCHAR('/') == *(lpszFrom + 1))
		lpszFrom++;
	if (--dwLen < cchTo) {	/* length without null character. */
		lpszTo += dwLen;
		cchTo -= dwLen;
	} else {
		lpszTo = NULL;
		cchTo = 0;
	}
	return(PosixlyToWin32(lpszFrom, lpszTo, cchTo) + dwLen);
}

/*
 *	DWORD FixPath(LPCTSTR, LPTSTR, DWORD, LPTSTR *, BOOL);
 *	fix the given path to Win32 full path and returns the length of the
 *	converted path.
 */
DWORD FixPath(LPCTSTR lpszFrom,
		LPTSTR lpszTo, DWORD cchTo, LPTSTR *ppszFile, BOOL bAllowCygwin)
{
	DWORD dwRet = 0;
	TCHAR szFix[_MAX_PATH], szExpand[_MAX_PATH];

	if (_TCHAR('/') == *lpszFrom && bAllowCygwin)
		dwRet = CygwinToWin32(lpszFrom, szFix, TSIZE(szFix));
	if (!dwRet) dwRet = PosixlyToWin32(lpszFrom, szFix, TSIZE(szFix));

	if (dwRet) {
		ExpandEnvironmentStrings(szFix, szExpand, TSIZE(szExpand));
		dwRet = GetFullPathName(szExpand, cchTo, lpszTo, ppszFile);
	}

	return(dwRet);
}
